;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                           Introduction
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;
;;; ernie is a grep-like script written in 100% Emacs Lisp.  It should
;;; just work.  The hope is that this will solve the problem of grepping
;;; under Windows, and that ernie will get picked up by the official emacs
;;; distribution and used as the fallback if a compiled grep cannot be
;;; found.  For inclusion with the main GNU Emacs source, permission is
;;; hereby given in advance to switch to any GPL license.
;;;
;;; Another advantage of ernie is that you only have to learn one regular
;;; expression syntax.  The same is basically true of the built-in grep
;;; front-ends which can now share the same regular expression syntax with
;;; the grep back-end.
;;; 
;;; The main disadvantage of ernie is that it is not as fast as a compiled
;;; grep, but it is infinitely faster than no grep.
;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                              Usage
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;
;;; For stand-alone operation, run this script as follows:
;;;
;;;     emacs -Q --batch --load ernie -f ernie/main \
;;;           -- [<options>] <regexp> <file> [<file> ... ]
;;;
;;; If you choose to start emacs directly, remember that the options
;;; for this script should be proceeded by "--" to prevent emacs from
;;; trying to consume them.
;;;
;;; Two types of wrappers are provided to help hide the details of the raw
;;; command above that is needed to start emacs in batch mode.  These
;;; wrappers are called "grep" and "ernie" respectively.  The "grep" type
;;; wrapper is what you would expect from any grep:
;;;
;;;     # Unix
;;;     grep.sh -r -n -i -e <regexp> <dir>
;;;
;;;     # Windows
;;;     grep.bat -r -n -i -e <regexp> <dir>
;;;
;;; But typing "grep -r -n -i e" at the command line is a little slow.
;;; So, the second type of wrapper is called "ernie" which gets its name
;;; from the most common way of starting grep:
;;;
;;;     grep -r -n -i -e
;;;       ^   ^  ^  ^  ^
;;;
;;; The "ernie" wrappers are equivalent to "grep -r -n -i e" so you would
;;; typically use them on the command line as follows:
;;;
;;;     ernie <regexp> <dir>
;;;
;;; Other options can follow like --include or --exclude, but the first
;;; positional parameter for ernie must be a regular expression.  For more
;;; control, use the "grep" wrapper.
;;;
;;; If you want the "ernie" type of wrapper to print the version, "ernie
;;; -V" will not work because the -V is in the "-e <regexp>", and the -e
;;; option says to allow leading dashes as part of the regexp.  To get
;;; ernie to print its version, do something like the following:
;;; 
;;;     ernie -- -V
;;; 
;;; Lastly, single quotes under Windows do not act like single quotes
;;; under bash.  If in doubt, use double quotes under Windows.
;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                           Installation
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;
;;; To install, just download the source and unpack somewhere convenient.
;;; If you use the included "ernie.bat" or "grep.bat" wrapper, it is not
;;; necessary for the directory containing "ernie.el" to be added to your
;;; EMACSLOADPATH.  It is helpful though to have both "emacs" and the
;;; provided wrappers in your system PATH.
;;;
;;; ernie is designed (within reason) to be a drop-in replacement for the
;;; grep command.  As such, this script is designed to be run as a
;;; separate program.  If you are having trouble shelling out under
;;; Windows, try using Emacs's "cmdproxy.exe" as your shell as follows:
;;;
;;;     ;; For Windows,
;;;     (setq explicit-shell-file-name
;;;           (concat exec-directory "cmdproxy.exe"))
;;;
;;; Then, update your ~/.emacs file to point to the "grep.bat" wrapper as
;;; follows:
;;;
;;;     (setq grep-command "grep.bat -H -r -n -i -e ")
;;;     (setq grep-find-command "grep.bat -H -r -n -i -e ")
;;;
;;; You will probably also want to add the following which is not
;;; necessary provided you use the -H option as shown above:
;;;
;;;     ;; Prevent emacs from silently adding NUL to the grep command line.
;;;     (setq grep-use-null-device nil)
;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                              Extras
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;
;;; The following command-line options are extras:
;;;
;;;                      --abspath : Print absolute paths.
;;;   --line-length-limit <length> : Truncate long lines.
;;;            --size-limit <size> : Skip large files.
;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Requires
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require 'cl)
(require 'regexp-opt)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Variables
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar ernie/version-info '(1 0 10)
  "Version in the form '(major minor patch).")

(defvar ernie/help-message "\
Usage:
=====

    emacs -Q --batch --load ernie.el -f ernie/main \\
          -- [<options>] <regexp> [<file> ...]

    (For an easier time, see the \"Examples\" subsection below
    for how to use the provided wrappers.)



Options:
=======

    --abspath

        By default the file name is prefixed to the matching
        lines of text as a relative path.  This option will
        prefix absolute paths instead.

    --A N
    --after-context N

        Print the N lines that come immediately after the match.
        This option takes precedence over the --context option.

    --B N
    --before-context N

        Print the N lines that come immediately before the match.
        This option takes precedence over the --context option.

    --C N
    --context N

        Print the N lines that come immediately before the match
        and the N lines to come immediatly after the match.  This
        option is overridden by the --after-context and
        --before-context options.

    -c
    --count

        Print the count of the number of matches found.  Do not
        print any matching lines.

    -e
    --regexp <regexp>

        Set the regular expression.  This option can be used more
        than once to express alternation.  If and only if this
        option is not set explicitly, the first positional
        parameter is used as the regular expression.

    --exclude <regexp>

        Any file matching the regular expression is excluded.
        This option can be used more than once to express
        alternation.  By default, no files are excluded.

    -F
    --fixed-strings

        Treat the search pattern as a fixed string instead of as
        a regular expression.

    -h
    --no-filename

        By default, when matching multiple files, the output
        lines are prefixed with the name of the file that
        contains the matching line.  This option prevents the
        file name from being printed.

    -H
    --with-filename

        By default, when matching a single file, the output lines
        are not prefixed with the name of the file that contains
        the matching line.  This option enables the file name to
        be printed.

    --help

        Print the help message and exit.

    -i
    --ignore-case

        Perform a case-insensitive search.

    --include <regexp>

        Only files matching the regular expression are included.
        This option can be used more than once to express
        alternation.  By default all files are included.

    --line-length-limit <length>

        Truncate each matching line to <length> characters.  By
        default or if size is 0, there is no limit on the line
        length.  The length can be specified in characters or
        with a suffix of \"K\", \"M\", or \"G\" to indicate
        kilo-characters, mega-characters, or giga-characters
        respectively.  Lower-case suffix letters are also valid.

    -n
    --line-number

        Prefix the matching line with the line number.

    -r
    -R
    --recursive

        Recursively grep directories.  By default, directories
        are ignored and not part of the grep.

    --size-limit <size>

        Limit the files that are grepped to those that are less
        than the specified size.  By default or if size is 0,
        there is no limit on the file size.  Size can be
        specified in bytes or with a suffix of \"K\", \"M\", or
        \"G\" to indicate kilobytes, megabytes, or gigabytes
        respectively.  Lower-case suffix letters are also valid.

    -v
    --invert-match

        Only print lines that do not match.

    -V
    --version

        Print the version and exit.  For the \"ernie\" wrapper,
        you need to provide a dummy regular expression (to match
        with the implicit trailing -e option):

            ernie - -V

    -w
    --word-regexp

        To match, the substring must start and end on a word boundary.



Examples:
========

  1) # Use \"--help\" to get help, not \"-h\"!
     emacs -Q --batch -L `pwd` --load ernie.el -f ernie/main -- --help

  2) # Make sure you remember the \"--\" to separate options.
     # Make sure you remember the \"-r\" to recursively search.
     emacs -Q --batch -L `pwd` --load ernie.el -f ernie/main \\
           -- -r -n -i -e foo .

  3) # On Unix, you can use the \"grep.sh\" wrapper to simplify.
     ./grep.sh -r -n -i -e foo .

  4) # On Windows, you can use the \"grep.bat\" wrapper to simplify.
     .\\grep.bat -r -n -i -e foo .

  5) # Both commands above can be shorted using the \"ernie\" wrapper.
     ernie foo .
")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro portable-labels (&rest forms)
  "Macro to deal with the deprecation of (labels) in favor
of (cl-labels) in the cl package."
  (if (fboundp 'cl-labels)
      `(cl-labels ,@forms)
    `(labels ,@forms)))

(defun interpose (xs y)
  "Interpose the Y element between each element of the XS list.

    (apply #'concat (-> \"PATH=foo;bar;baz\"
                        (split-string \";\")
                        (interpose \":\"))))
"
  (let (rv)
    (when (consp xs)
      (setq rv (cons (first xs) rv))
      (dolist (x (rest xs))
        (setq rv (cons x (cons y rv)))))
    (reverse rv)))

(defun ernie/make-default-options ()
  "Return a hash table with each valid option set to its default value."
  (let ((rv (make-hash-table :test 'equal)))
    (puthash :abspath nil rv)
    (puthash :after-context nil rv)
    (puthash :before-context nil rv)
    (puthash :count nil rv)
    (puthash :exclude nil rv)
    (puthash :files nil rv)
    (puthash :fixed-strings nil rv)
    (puthash :ignore-case nil rv)
    (puthash :include nil rv)
    (puthash :invert-match nil rv)
    (puthash :line-number nil rv)
    (puthash :line-length-limit 0 rv)
    (puthash :no-filename nil rv)
    (puthash :recursive nil rv)
    (puthash :regexps nil rv)
    (puthash :size-limit 0 rv)
    (puthash :user-wants-help nil rv)
    (puthash :user-wants-version nil rv)
    (puthash :with-filename nil rv)
    (puthash :word-regexp nil rv)
    rv))

(defun ernie/print-hash-table (ht)
  (let ((acc))
    ;; Generate a list of strings of the form "key -> value" and save
    ;; them in the accumulator "acc".
    (maphash (lambda (k v) (setq acc (cons (format "%s -> %s" k v) acc))) ht)
    ;; Sort the strings and print them.  The one complicating factor
    ;; is that the keys have to be left justified in order to sort
    ;; correctly, but need to be right justified in order to print
    ;; correctly.  So, (string-match) and (match-string) are used to
    ;; extract the key/value pair and reformat them after they have
    ;; been sorted.
    (map nil
         (lambda (s)
           (string-match "\\([-a-zA-Z:]+\\) -> \\(.*\\)$" s)
           (princ (format "%24s -> %s\n"
                          (match-string 1 s)
                          (match-string 2 s))))
         (sort acc #'string<))))

(defun ernie/string-to-bytes (s)
"Convert the string s with byte-related suffixes like \"K\"
\(kilobytes\), \"M\" \(megabytes\), and \"G\" \(gigabytes\) to
the equivalent number of bytes as a floating point value.  Thus,
\"1K\" convert to 1024.0, \"1M\" to 1048576.0, and \"1G\" to
1073741824.0.  Floating point is used because Emacs Lisp returns
file sizes in floating point when they are too large to be held
in native integer values."
  (if (not (string-match "^\\(\\+?[0-9]+\\)\\([kKmMgG]?\\)$" s))
      (error "Unable to convert string to size: %s" s))
  (let ((num (string-to-number (match-string 1 s)))
        (suffix (match-string 2 s)))
    (cond ((member suffix '("")) (float num))
          ((member suffix '("k" "K")) (* 1024.0 num))
          ((member suffix '("m" "M")) (* 1024.0 1024.0 num))
          ((member suffix '("g" "G")) (* 1024.0 1024.0 1024.0 num))
          (t (error "Internal Error: ernie/string-to-bytes: %s" s)))))

(defmacro ernie/next-arg (index arg args argc)
  "Increment INDEX to the next index value and advance ARG to the
next argument in ARGS.  If INDEX would be advanced beyond the end
of ARGS (as denoted by the argument count ARGC), an error
condition is signaled."
  `(progn
     (if (>= (+ ,index 1) ,argc)
         (error "Missing value for argument: %s" ,arg))
     (setq ,index (+ ,index 1))
     (setq ,arg (elt ,args ,index))))

(defun ernie/post-process-the-h-options (opts)
  "The -h and -H options so if one is set, the other should be
cleared."
  ;;
  ;; Whether to print file names is historically complicated but
  ;; provides reasonable defaults.  By default, do not print file
  ;; names when only one file is specified but do print file names
  ;; by default when more than on file is specified.  (This is
  ;; helpful when scripting because it saves you from having to
  ;; strip off the file name in this common case.)
  ;;
  ;; So, the only time your really need to use the -H option is when
  ;; one file is specified but for some reason you want the file name
  ;; printed as part of the output.  Likewise, the only time you
  ;; really need to specify the -h option is when more than one file
  ;; is specified or if the -r (--recursive) option is specified and
  ;; for some reason you do not want the file names printed as part of
  ;; the output.
  ;;
  ;; NOTE: The command-line convention is for -H to override -h, but
  ;; the code for this program is written in terms of -h.  So, if -H
  ;; is set, this code requires -h to be cleared.
  ;;

  ;; If neither option (-h or -H) was specified on the command line,
  ;; the default value depends on how many files were specified.
  (if (and (not (gethash :no-filename opts))
           (not (gethash :with-filename opts)))
      ;; Neither was set.
      (cond ((gethash :recursive opts)
             (puthash :with-filename t opts))
            ((<= (length (gethash :files opts)) 1)
             (puthash :no-filename t opts))
            (t (puthash :with-filename t opts)))
    ;; At least one was set.  If -H is set, it overrides -h.
    (if (gethash :with-filename opts)
        (puthash :no-filename nil opts))))

(defun ernie/parse-command-line (args)
  "Parse the command-line options in ARGS and return the results
as a hash table."
  (let ((argc (length args))
        (rv (ernie/make-default-options)))

    ;; Iterate over the command-line options.
    (do ((i 0 (setq i (+ i 1))))
        ((>= i argc))
      (let ((arg (elt args i)))
        (cond
         ;; Skip empty argument strings.
         ((string= arg "") nil)
         ;; Handle all the command-line arguments that start with a hyphen.
         ((member arg '("--abspath"))
          (puthash :abspath t rv))
         ((member arg '("-A" "--after-context"))
          (ernie/next-arg i arg args argc)
          (puthash :after-context (string-to-number arg) rv))
         ((member arg '("-B" "--before-context"))
          (ernie/next-arg i arg args argc)
          (puthash :before-context (string-to-number arg) rv))
         ((member arg '("-C" "--context"))
          (ernie/next-arg i arg args argc)
          (unless (gethash :after-context rv)
            (puthash :after-context (string-to-number arg) rv))
          (unless (gethash :before-context rv)
            (puthash :before-context (string-to-number arg) rv)))
         ((member arg '("-c" "--count"))
          (puthash :count t rv))
         ((member arg '("-e" "--regexp"))
          (ernie/next-arg i arg args argc)
          (puthash :regexps (cons arg (gethash :regexps rv)) rv))
         ((member arg '("--exclude"))
          (ernie/next-arg i arg args argc)
          (puthash :exclude (cons arg (gethash :exclude rv)) rv))
         ((member arg '("-F" "--fixed-strings"))
          (puthash :fixed-strings t rv))
         ((member arg '("-h" "--no-filename"))
          (puthash :no-filename t rv))
         ((member arg '("-H" "--with-filename"))
          (puthash :with-filename t rv))
         ((member arg '("--help"))
          (puthash :user-wants-help t rv))
         ((member arg '("-i" "--ignore-case"))
          (puthash :ignore-case t rv))
         ((member arg '("--include"))
          (ernie/next-arg i arg args argc)
          (puthash :include (cons arg (gethash :include rv)) rv))
         ((member arg '("--line-length-limit"))
          (ernie/next-arg i arg args argc)
          (let ((len (ernie/string-to-bytes arg)))
            (condition-case nil
                (setq len (truncate len))
              (error (error "Line length limit out of range: %s" arg)))
            (puthash :line-length-limit len rv)))
         ((member arg '("-n" "--line-number"))
          (puthash :line-number t rv))
         ((member arg '("-r" "-R" "--recursive"))
          (puthash :recursive t rv))
         ((member arg '("--size-limit"))
          (ernie/next-arg i arg args argc)
          (puthash :size-limit (ernie/string-to-bytes arg) rv))
         ((member arg '("-v" "--invert-match"))
          (puthash :invert-match t rv))
         ((member arg '("-V" "--version"))
          (puthash :user-wants-version t rv))
         ((member arg '("-w" "--word-regexp"))
          (puthash :word-regexp t rv))
         ((member arg '("-"))
          (puthash :files (cons arg (gethash :files rv)) rv))
         ;; Anything else that starts with a hyphen is an invalid argument.
         ((= (elt arg 0) ?-)
          (error "Unexpected option: %s" arg))
         ;; For now, treat all of the positional parameters as files
         ;; or directories to be grepped (but see below).
         (t (puthash :files (cons arg (gethash :files rv)) rv)))))

    ;;
    ;; Post-process the options.
    ;;

    ;; Reverse the list of files to get the correct ordering.
    (puthash :files (reverse (gethash :files rv)) rv)

    ;; If the pattern was not set with a -e or --regexp option, the
    ;; first positional parameter is the pattern.
    (unless (gethash :regexps rv)
      (let ((files (gethash :files rv)))
        ;; Note that we do not signal an error condition here even if
        ;; the :regexps list is empty because options like --help do
        ;; not require any patterns.  Instead, this type of error has
        ;; to be signaled in the main program.
        (when files
          (puthash :regexps (list (first files)) rv)
          (puthash :files (rest files) rv))))

    ;; Straighten out the -h (--no-filename) and -H (--with-filename) options.
    (ernie/post-process-the-h-options rv)

    ;; Reverse the remaining lists.
    (dolist (sym '(:exclude :include :regexps))
      (puthash sym (reverse (gethash sym rv)) rv))

    ;; If no --include regexp was specified, the default is to include
    ;; all files.  So, the rest of the code will be easier if the list
    ;; of files to include defaults to the empty regular expression
    ;; (which matches all strings).
    (unless (gethash :include rv)
      (puthash :include '("") rv))

    ;; If :after-context was not set above, default it to 0 lines.
    (unless (gethash :after-context rv)
      (puthash :after-context 0 rv))
    (if (minusp (gethash :after-context rv))
        (error "\"after\" context is negative: %d"
               (gethash :after-context rv)))

    ;; If :before-context was not set above, default it to 0 lines.
    (unless (gethash :before-context rv)
      (puthash :before-context 0 rv))
    (if (minusp (gethash :before-context rv))
        (error "\"before\" context is negative: %d"
               (gethash :before-context rv)))

    ;; Return the hash.
    rv))

(defun ernie/combine-alternative-regexps (rs)
  "Combine the list of alternative regular expressions RS into
one equivalent regular expression and return the result.  If RS
is empty, nil is returned."
  (let ((len (length rs)))
    (cond ((zerop len) nil)
          ((= len 1) (car rs))
          (t (format "\\(?:%s\\)"
                     (apply #'concat
                            (interpose rs "\\)\\|\\(?:")))))))

(defun ernie/anchor-regexp-at-word-boundaries (regexp)
  "Anchor the regular expression REGEXP at word boundaries such
that any matching substring has to start and end on a word
boundary."
  ;; A "shy group" is used to avoid changing the number of groups for
  ;; the purpose of capture.
  (format "\\<\\(?:%s\\)\\>" regexp))

(defun ernie/print-matching-line (line line-count sep opts fname)
  "Print the matching line LINE given the user's options OPTS and
 the file name FNAME from which the line was read.  The separator
 SEP is used to separate logical units on the line.  This
 function is called for both normal or inverted matches because
 the format of the output is the same in both cases."
  (let ((count (gethash :count opts))
        (no-filename (gethash :no-filename opts))
        (line-number (gethash :line-number opts))
        (line-length-limit (gethash :line-length-limit opts)))
    (unless count
      (princ (concat (if (not no-filename) (concat fname sep))
                     (if line-number (format "%d%s" line-count sep))
                     (if (and (plusp line-length-limit)
                              (> (length line) line-length-limit))
                         (substring line 0 line-length-limit)
                       line)
                     "\n")))))

(defun ernie/print-final-match-count (match-count opts fname)
  "Print the final count of input strings that matched the
regular expression."
  (let ((count (gethash :count opts))
        (no-filename (gethash :no-filename opts)))
    (if count
        (princ (format "%s%d\n"
                       (if no-filename "" (concat fname ":"))
                       match-count)))))

(defun* ernie/split-into-lines (text &key
                                     (func nil)
                                     (initial-value nil))
  "Split the TEXT into lines and, by default, return the list of
lines.  If FUNC is provided, it will be passed three parameters
with one of the parameters being an accumulator similar to the
accumulator used by (reduce).  The first parameter is the next
line of text.  The second parameter is the line number which is
one-based so it will be 1 the first time FUNC is called.  The
third parameter is the INITIAL-VALUE the first time FUNC is
called.  For subsequent calls, it is whatever FUNC returned on
the previous call.  The INITIAL-VALUE is only used if FUNC is
defined.  FUNC should return two values.  The first value should
be true to indicate that callbacks should continue and be false
to indicate callbacks should stop.  The second value FUNC should
return is the (reduce)-like accumulator.

This function is different from Emacs Lisp's (split-string)
function because it can provide immediate feedback to the user if
FUNC is passed in instead of having to wait for the entire list
to be generated before starting to work on it.  It also
understands that files without a final EOL sequence do not have
one less line than files with a final EOL sequence."
  ;; Iterate over each match.
  (let ((done nil)
        (index 0)
        (start 0)
        (stop 0)
        (acc (if func initial-value nil)))
    (do ((line-count 1 (setq line-count (+ line-count 1))))
        ((or done (> index stop))
         ;; Return the user's accumulator as is, or our accumulator after
         ;; reversing the list of lines to put them in the correct order.
         (if func acc (reverse acc)))
      ;; Find the next line of text.
      (when (string-match "\r?\n\\|$" text index)
        (let ((mdata (match-data)))
          (setq start (first mdata))
          (setq stop (second mdata))
          (let ((line (substring text index start)))
            ;; Advance our index to the start of the next line.  The
            ;; purpose of the (if) is to nudge "index" passed the "$"
            ;; match because the Emacs Lisp regexp engine seems to get
            ;; stuck there.
            (setq index (+ stop (if (= index stop) 1 0)))
            (unless (> index stop)
              (multiple-value-bind (wants-more acc*)
                  (if func (funcall func line line-count acc)
                    (values t (funcall #'cons line acc)))
                (setq done (not wants-more))
                (setq acc acc*)))))))))

(defun ernie/print-line-context-ring (line-context-ring opts fname)
  "The context ring is a circular buffer that forms a sliding
window over the text being grepped.  That windows is strictly
ordered with the lines comprising the \"before\" context,
followed by the \"line in context\", followed by the lines
comprising the \"after\" context.  If the \"line in context\"
matched (as determined by the data stored along with the line in
the LINE-CONTEXT-RING), this method prints the entire context;
otherwise, it prints nothing.  Lines that are NUL are not
printed."
  ;; There is no "line in context" until the line-context-ring is
  ;; full.  Remember that ":before-context"-worth of NILs are
  ;; initially added to the line-context-ring such that the first real
  ;; line added to the line-context-ring falls into the main slot.
  (when (= (ring-length line-context-ring)
           (ring-size line-context-ring))
    (let*
        ;; Unpack the options.
        ((invert-match (gethash :invert-match opts))
         (case-fold-search (gethash :ignore-case opts))
         ;; Note that the correct index for the main line is between
         ;; the "before" context and the "after" context.  In other
         ;; words, it immediately follows the "before" context.
         (main-index (gethash :before-context opts))
         ;; If the entire length of the line-context-ring is just 1,
         ;; there is obviously no context.
         (has-no-context (<= (ring-length line-context-ring) 1))
         (main-data (ring-ref line-context-ring main-index)))

      ;; Carefully get the data for the main line (aka the "line in context").
      (when main-data
        (destructuring-bind
            (main-matched match-count main-line-count main-line) main-data
          (when main-matched
            ;; Print the context block separator if this isn't the
            ;; first match or if no context was specified.  The reason
            ;; for first-match is so that the separator is only
            ;; present *between* context blocks.  We also don't want
            ;; to print the separator if the "--count" option is
            ;; given.
            (unless (or (<= match-count 1) has-no-context (gethash :count opts))
              (princ "--\n"))
            ;; Iterate over the lines of context (including the main line).
            (do ((i 0 (setq i (+ i 1)))
                 (total (ring-length line-context-ring)))
                ((>= i total))
              ;; Carefully get the next line.
              (let ((next-data (ring-ref line-context-ring i)))
                (when next-data
                  (destructuring-bind
                      (next-matched _ next-line-count next-line) next-data
                    ;; Print the next line.
                    (ernie/print-matching-line next-line
                                               next-line-count
                                               (if (= i main-index) ":" "-")
                                               opts
                                               fname)))))))))))

(defun ernie/flush-line-context-ring (line-context-ring opts fname)
  "After the main loop exits, there are still unprocessed entries
in the LINE-CONTEXT-RING which need to be flushed.  This function
flushes those entries.  Those entries haven't been flushed yet
because there is a delay in the LINE-CONTEXT-RING that allows
that part of the code to run in linear time (instead of quadratic
time)."
  ;; Get the main index.
  (let ((main-index (gethash :before-context opts))
        ;; Determine if the context ring at the time when this
        ;; function was called.
        (was-full (= (ring-length line-context-ring)
                     (ring-size line-context-ring))))
  ;; If the line-context-ring is full, flush the oldest line because
  ;; the main loop has already processed it fully, but if the
  ;; line-context-ring is not full, it means we need to insert enough
  ;; NILs to fill it so that we bump each line forward by inserting
  ;; just one more NIL.
  (if was-full
      ;; Flush the last line that was processed by the main loop.
      (ring-insert-at-beginning line-context-ring nil)
    ;; Else, insert just enough NILs to fill the ring.
    (while (< (ring-length line-context-ring)
              (ring-size line-context-ring))
      (ring-insert-at-beginning line-context-ring nil)))
  ;; Iterate until all lines have passed through the main index.
  (while (ring-ref line-context-ring main-index)
    ;; Print this line.
    (ernie/print-line-context-ring line-context-ring opts fname)
    ;; Flush the oldest line.
    (ring-insert-at-beginning line-context-ring nil))))

(defun ernie/grep-line/make-client-data (fname opts)
  "Use the file name FNAME and options OPTS to initialize the
client data needed for the (ernie/grep-line) callback.  The
resulting hash that can be used as the client data
for (ernie/grep-line) is returned."
  (let* ((client-data (make-hash-table :test 'equal))
         (after-context (gethash :after-context opts))
         (before-context (gethash :before-context opts))
         (word-regexp (gethash :word-regexp opts))
         (total-lines (+ before-context 1 after-context)))

    (puthash :opts opts client-data)
    (puthash :match-count 0 client-data)
    (puthash :file-name fname client-data)

    (puthash :users-regexp
             (cond
              ;; Convert multiple fixed strings to one regexp.
              ;; This lets us use the same regexp engine giving
              ;; us the same options for fixed strings as for
              ;; regexps.
              ((gethash :fixed-strings opts)
               (regexp-opt (gethash :regexps opts)))
              (t
               ;; Convert multiple, smaller alternative regexps
               ;; to one, larger regexp.
               (ernie/combine-alternative-regexps
                (gethash :regexps opts))))
             client-data)

    ;; This should be the last manipulation of :users-regexp because
    ;; it needs to wraps the entire regexp in "\\<...\\>".
    (if word-regexp
        (puthash :users-regexp
                 (ernie/anchor-regexp-at-word-boundaries
                  (gethash :users-regexp client-data))
                 client-data))

    ;; In both cases we are using (string-match) to do regular
    ;; expression matching because above for :users-regexp we
    ;; convert the fixed strings to a regular expression using
    ;; (regexp-opt).
    (puthash :search-func
             (cond ((gethash :fixed-strings opts) #'string-match)
                   (t #'string-match))
             client-data)

    ;; Create the line-context-ring.  This holds the "before"
    ;; context, the main line, and then the "after" context.
    (let ((line-context-ring (make-ring total-lines)))
      (puthash :line-context-ring line-context-ring client-data)

      ;; The line-context-ring needs us to push in exactly enough
      ;; NIL values to fill the "before" part of the context.
      ;; This allows the first line inserted to go into the "main"
      ;; slot which is the only slot from which matches will be
      ;; printed.
      (dotimes (_ before-context)
        (ring-insert-at-beginning line-context-ring nil)))

    ;; Return the client data.
    client-data))

(defun ernie/grep-line (line line-count client-data)
  "This is a callback function invoked by (ernie/split-into-lines).
It is called once for each LINE.  The line number for LINE should
be passed in as LINE-COUNT.  The CLIENT-DATA is the data passed
in when this function was registered."
  (let*
      ;; Unpack the client data.
      ((opts (gethash :opts client-data))
       (invert-match (gethash :invert-match opts))
       (case-fold-search (gethash :ignore-case opts))
       (fname (gethash :file-name client-data))
       (users-regexp (gethash :users-regexp client-data))
       (search-func (gethash :search-func client-data))
       (line-context-ring (gethash :line-context-ring client-data))
        ;; Try to match the user's regexp against the line.
       (preliminary-match (funcall search-func users-regexp line)))

    ;; Carefully increment the total match count taking into account
    ;; the "--invert-match" ("-v") option if specified.
    (let ((line-matched (or (and preliminary-match (not invert-match))
                            (and (not preliminary-match) invert-match))))
      (if line-matched
          ;; Increment the match count.
          (let ((match-count (+ 1 (gethash :match-count client-data))))
            (puthash :match-count match-count client-data)))

      ;; Insert the current line into the circular buffer.  It is the
      ;; circular buffer that allows -A, -B, and -C to work
      ;; efficiently (in linear time).  The main problem is the
      ;; "after" context because we haven't seen those lines yet.  If
      ;; we immediately go find its "after" context, we get quadratic
      ;; behavior.  Better is to just queue the line in a circular
      ;; buffer and test the one line in the circular buffer that has
      ;; a full "before" and "after" context.  (The size of the
      ;; circular buffer is set by the caller to be (+ before-context
      ;; 1 after-context).)  The main problem with this approach is
      ;; that there will be unprocessed lines remaining in the
      ;; circular buffer after EOF is reached so we have to arrange to
      ;; have the caller flush these lines.
      ;;
      ;; Also note that the caller initially pushed in enough NIL
      ;; values to completely fill the "before" context part of the
      ;; buffer.  This way the first line we insert goes into the
      ;; "main" slot in the circular buffer which is the only slot
      ;; from which (ernie/print-line-context-ring) will print a match.
      (ring-insert-at-beginning line-context-ring
                                (list line-matched
                                      (if line-matched
                                          (gethash :match-count client-data))
                                      line-count
                                      line))

      ;; Print the line-context-ring.  This does not necessarily print
      ;; the current line.  Instead, it prints the line that has been
      ;; delayed exactly long enough to be the "line in context."  See
      ;; comments for (ernie/print-line-context-ring) for more.
      (ernie/print-line-context-ring line-context-ring opts fname)))

  ;; Return t to (ernie/split-into-lines) so it will continue calling
  ;; this function for each line, and return client-data as the
  ;; "accumulator".
  (values t client-data))

(defun ernie/grep-standard-input (fname opts)
  "Grep the date read from standard input.  For each line in
standard input that matches any of the regular expressions
in (gethash :regexps OPTS), that line is printed to standard
output prefixed by the file name FNAME and line number where the
match occurs.  Here, the file name FNAME is just the name you use
to indicate standard input as the source of the data."

  ;; Initialize the client data used for (ernie/split-into-lines) callback.
  (let ((client-data (ernie/grep-line/make-client-data fname opts)))

    (condition-case err

        ;; Iterate over each line of input.
        (do ((line (read-string "") (read-string ""))
             (line-count 1 (+ 1 line-count)))
            ((not line))
          ;; Grep each line and accumulate the results.
          (ernie/grep-line line line-count client-data))

      ;; The source shows that (read-string) signals the generic
      ;; "error" condition so it is hard to distinguish between this
      ;; error condition and any other.  Try comparing against the
      ;; hard-coded string.  This condition is signaled on normal EOF
      ;; so something needs to be done with it.
      (error (if (not (string= (error-message-string err)
                               "Error reading from stdin"))
                 (error "%s" (error-message-string err)))))

    ;; To run in linear time, there is a lag in the
    ;; line-context-ring.  Flush the remaining lines.
    (let ((line-context-ring (gethash :line-context-ring client-data)))
      (ernie/flush-line-context-ring line-context-ring opts fname))

    ;;  Print the final match count if required.
    (let ((match-count (gethash :match-count client-data)))
      (ernie/print-final-match-count match-count opts fname))))

(defun ernie/grep-buffer (fname opts buffer)
  "Grep the BUFFER loaded from file FNAME.  For each line in
BUFFER that matches any of the regular expressions
in (gethash :regexps OPTS), that line is printed to standard
output prefixed by the file name FNAME and line number where the
match occurs."
  (with-current-buffer buffer
    (save-excursion
      (save-restriction
        (widen)

        ;; Move to the beginning of the buffer.
        (goto-char (point-min))

        ;; Initialize the client data used for
        ;; (ernie/split-into-lines) callback.
        (let ((client-data (ernie/grep-line/make-client-data fname opts)))

          ;; Grep each line and accumulate the results.
          (ernie/split-into-lines (buffer-string)
                                  :func #'ernie/grep-line
                                  :initial-value client-data)

          ;; To run in linear time, there is a lag in the
          ;; line-context-ring.  Flush the remaining lines.
          (let ((line-context-ring (gethash :line-context-ring client-data)))
            (ernie/flush-line-context-ring line-context-ring opts fname))

          ;;  Print the final match count if required.
          (let ((match-count (gethash :match-count client-data)))
            (ernie/print-final-match-count match-count opts fname)))))))

(defun ernie/grep-file (fname opts)
  "Grep the file FNAME.  For each line in FNAME that matches any
of the regular expressions in (gethash :regexps OPTS), that line
is printed to standard output prefixed by the file name and line
number where the match occurs."
  ;; Make sure the file isn't too large.
  (condition-case err
      (let ((size-limit (gethash :size-limit opts))
            (attrs (file-attributes fname)))
        (if (not attrs)
            (signal 'file-error
                    (list (format "Cannot stat file: %s" fname)))
          (when (or (zerop size-limit)
                    (let ((actual-size (nth 7 attrs)))
                      (< actual-size size-limit)))

            ;; Make sure the file passes the --include and --exclude filters.
            (let ((incl (gethash :include opts))
                  (excl (gethash :exclude opts))
                  (case-fold-search nil))
              ;; We only grep "files" that are regular files, match at
              ;; least one --include regular expressions, and don't
              ;; match any --exclude regular expressions.  Note that
              ;; the empty regular expression (which matches all) is
              ;; added to the :include list by default, i.e., only if
              ;; the user doesn't specify an :include list.
              (when (and (file-regular-p fname)
                         (some (lambda (regexp)
                                 (string-match  regexp fname))
                               incl)
                         (notany (lambda (regexp)
                                   (string-match regexp fname))
                                 excl))

                ;; Replace the *scratch* buffer with the literal
                ;; contents of the file.
                (insert-file-contents-literally fname  ; filename
                                                t      ; visit
                                                nil    ; begin
                                                nil    ; end
                                                t)     ; replace
                ;; Grep the buffer.
                (ernie/grep-buffer fname opts (current-buffer)))))))

    ;; Error Conditions
    (error
     (message "*** Warning: %s" (error-message-string err)))))

(defun ernie/grep-file* (fname opts)
  "This function is a wrapper around (ernie/grep-file).  It is is
designed to be used with (ernie/dirtree) and is only called with
the user specifies --recursive.  It is used to filter in files
based on the --include regular expressions and to filter out
files based on the --exclude regular expressions both in OPTS."
  ;; Grep the file.
  (ernie/grep-file fname opts)
  ;; Return "opts" so it will be passed in on the next callback.
  opts)

(defun* ernie/dirtree (dirs &key
                            (abspath nil)
                            (attrs nil)
                            (depth nil)
                            (id-format 'string)
                            (func nil)
                            (initial-value nil))
  "Return the recursively generated list of files that reside in
the list of directories DIRS.

If ABSPATH is true, absolute paths will be returned; otherwise,
paths relative to the directories in DIR are returned.

If ATTRS is true, the directory tree that is returned is a
directory tree of attributes (which includes the file names);
otherwise, just the file names are returned.
See (directory-files-and-attributes) for a description of the
attributes returned.

The ID-FORMAT can be either 'STRING or 'INTEGER depending on
whether the user ID and group ID should be listed as a string or
an integer.  This parameter only effective when ATTRS is not NIL.

By default, a directory is added to the list before recursively
adding its contents, but if DEPTH is true, the directory is added
to the list after its contents.

The behavior described above can be tailored by the caller by
providing the function FUNC which will be passed two parameters
similar to (reduce).  The first parameter is the file name or
file attributes (depending on whether ATTRS is set) for the
current file.  The second parameter is the INITIAL-VALUE for the
first call, and for subsequent calls, it is whatever was returned
by FUNC on the previous call.  If INITIAL-VALUE is used without
changing the default FUNC, the INITIAL-VALUE must be a list, and
it will be reversed before the first call, and the final
accumulator will be reversed again after the last call.  Note
that you can disable the extra calls to reverse but otherwise
maintain the default behavior by just passing in \":func #'cons\"
as the callback function.

Unconventionally, directory names are returned with a trailing
dot (e.g., \"foo/.\").  This is because this is how Emacs Lisp
returns the names and against which patterns like \"^\\\\.\"
match.  So for consistency, this is how the directory names are
passed through to the user.

One directory can be passed in as just a string, but multiple
directories must be passed in as a list of strings as follows:

    (ernie/dirtree (list dir1 dir2 dir3))

The following is an example of how to list all the files on the
file system:

    (ernie/dirtree \"/\" :func (lambda (fname acc) (message \"%s\" fname) nil))

The same command can be invoked directly from the command line as
follows:

    emacs -Q --batch --load ~/.emacs --eval \"(ernie/dirtree \\\"/\\\" :func (lambda (fname acc) (princ (format \\\"%s\\n\\\" fname)) nil))\"

The above code samples are designed to be read with

    (describe-function 'ernie/dirtree)
"
  (let
      ;; Keep a hash table of state needed to avoid infinite recursion
      ;; if symbolic links on the file system for a cycle.  The keys
      ;; for the "visited" hash table are the
      ;; (file-truename) values for the directories that have been
      ;; visited.  This hash table is checked before trying to
      ;; recursively descend into a directory to avoid infinite loops
      ;; created by symbolic links.
      ((visited (make-hash-table :test 'equal)))

    (portable-labels (

      ;; Return whether FNAME is the "." directory.
      (is-dot (fname)
        (string= (file-name-nondirectory fname) "."))

      ;; Return whether FNAME is the ".." directory.
      (is-dot-dot (fname)
        (string= (file-name-nondirectory fname) ".."))

      ;; Get the "realpath" (or what Lisp calls the "truename") for
      ;; the file FNAME.  If an error occurs, nil is returned.  This
      ;; function can fail to resolve the file name including but not
      ;; limited to race conditions that cause the file to already be
      ;; removed and cycles/infinite loops created by symbolic links.
      (get-realpath (fname)
        (condition-case e
            (file-truename fname)
          (error (message "%s" (error-message-string e))
                 nil)))

      ;; Return the list of files in the directory DIR.  By default,
      ;; the files are prepended to the accumulator ACC by the (cons)
      ;; function FUNC.
      (inner (dir func acc)

         ;; Mark the current directory as visited.  The call to
         ;; (get-realpath) can return an error so we have to be
         ;; careful about not continuing in that case.  One way an
         ;; error can occur is if the file has been removed in the
         ;; time between when we list it as part of its parent
         ;; directory and when we call (get-realpath).
         (let ((dir-key (get-realpath dir)))
           (when dir-key
             (puthash dir-key t visited)

             ;; Be able to remember the current directory DIR in case
             ;; DEPTH is set which requires us to add DIR to the list
             ;; after recursively processing all of its contents.
             (let ((current-dir))

               ;; Carefully iterate over the files in this directory.
               ;; An error can occur, for example, if you don't have
               ;; permission to list the contents of a directory.
               (dolist (fname-or-attrs
                 (condition-case e
                     ;; We can presumably save some time by not
                     ;; returning the full attribute list if the
                     ;; user only wants the file names.
                     (if attrs
                         (directory-files-and-attributes
                          dir abspath nil nil id-format)
                         (directory-files dir abspath))
                   (error (message "%s" (error-message-string e))
                          nil)))

                 ;; When (directory-files) does not return absolute
                 ;; paths, it returns just the base name so we have to
                 ;; join the base name with the directory name.
                 (when (not abspath)
                   (if attrs
                       ;; Replace the first element of the attrs
                       ;; (which is the base file name) with the
                       ;; relative file name.
                       (setq fname-or-attrs
                             (cons (concat (file-name-as-directory dir)
                                           (first fname-or-attrs))
                                   (rest fname-or-attrs)))
                     ;; Replace the base file name with the relative
                     ;; file name.
                     (setq fname-or-attrs
                           (concat (file-name-as-directory dir)
                                   fname-or-attrs))))

                 ;; Get the file name and whether it is a directory
                 ;; and whether this directory forms an infinite loop.
                 ;;
                 ;; "insert-before" determines whether to insert
                 ;; before or after inserting sub-directories.  It
                 ;; applies to files, symbolic links to directories
                 ;; that cause loops, and directories (including
                 ;; symbolic links to directories that do not cause
                 ;; loops).  The rules for files and symbolic links to
                 ;; directories that cause loops is trivial because
                 ;; neither case requires us to recursively process
                 ;; the contents.  For directories, it is a little
                 ;; more complicated because directories are typically
                 ;; seen twice.  Once in the parent directory, and
                 ;; then once again as the "."  of the current
                 ;; directory.  Because the top-level directories do
                 ;; not have parents, the "."  entry for the current
                 ;; directory is the canonical entry for a directory
                 ;; and will be the only one that is added to the list
                 ;; using the rule is that it gets added to the list
                 ;; before its contents is recursively added by
                 ;; default (i.e., if "depth" is false); otherwise
                 ;; (i.e., if "depth" is true), the current directory
                 ;; is added after recursively adding its contents.
                 (let* ((fname (if attrs (first fname-or-attrs) fname-or-attrs))
                        (is-dir (file-directory-p fname))
                        (is-current-dir (if is-dir (is-dot fname)))
                        (fname-key (get-realpath fname))
                        (is-loop (if (and is-dir (not is-current-dir))
                                     (gethash fname-key visited)))
                        (insert-before (or (not is-dir) is-loop
                                           (and (not depth) is-current-dir))))

                   ;; Only skip the ".." (parent) directory entry and
                   ;; directory entries where (file-truename) signaled
                   ;; a condition (causing "fname-key" to be nil).
                   (unless (or (is-dot-dot fname) (not fname-key))

                     ;; Remember the current directory but only if it
                     ;; needs to be inserted after recursively
                     ;; inserting its contents.
                     (when (and is-current-dir (not insert-before))
                       (setq current-dir fname-or-attrs))

                     (when insert-before
                       ;; Call FUNC with next file name (or file
                       ;; attributes) and the accumulator.
                       (setq acc (funcall func fname-or-attrs acc)))

                     ;; For directories, only recursively descend into
                     ;; them if they do not complete an infinite loop
                     ;; of symbolic links or if it is the current
                     ;; directory (over which we are currently
                     ;; iterating).
                     (when (and is-dir (not is-loop) (not is-current-dir))
                       ;; Recursively descend into the directory.
                       (setq acc (inner fname func acc))))))

               ;;
               ;; You reach here after recursively processing all of the
               ;; contents for the current directory.
               ;;

               ;; If DEPTH is specified, add the current directory now
               ;; (after having added all of its contents above).
               (when current-dir
                 ;; Call FUNC with next file name (or file attributes)
                 ;; and the accumulator.
                 (setq acc (funcall func current-dir acc))))))

        acc)

      ;; Return the list of files in all the directories in the list
      ;; of directories DIRS.  By default, the files are prepended to
      ;; the accumulator ACC by the (cons) function FUNC.
      (outer (dirs func acc)
        (dolist (dir dirs)
          (when (file-directory-p dir)
            ;; Reset the "visited" hash for each directory
            ;; specified by the user.  Presumably the user wants
            ;; the top-level directories fully expanded.
            (clrhash visited)
            ;; Recursively descend into the directory.
            (setq acc (inner dir func acc))))
        acc)
      )

      ;; If the user passes in DIRS as just a string, put it inside of
      ;; a list for the user.  Also, we need to reverse the
      ;; accumulator if the default FUNC (i.e. (cons)) is being used.
      ;; Also, if the user did not specify a value for FUNC, use the
      ;; default which is to just (cons) the next file name (or file
      ;; attributes) to the accumulator.
      (let* ((dirs (if (listp dirs) dirs (list dirs)))
             (is-reverse-needed (eq func nil))
             (func (if func func #'cons))
             (initial-value (if is-reverse-needed
                                (reverse initial-value)
                              initial-value)))
        ;; Generate the list of files.
        (let ((rv (outer dirs func initial-value)))
          (if is-reverse-needed (reverse rv) rv))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Main
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ernie/main ()
  "Entry point."

  (let ((error-level 0)
        ;; Silence some unused emacs features.
        (vc-handled-backends nil)
        (dired-use-ls-dired nil))

    (condition-case err
        (progn

        ;; Skipping the leading "--" option (if any) which is used to
        ;; prevent emacs from trying to parse our command-line options.
        (unless (and argv (not (string= (first argv) "--")))
          (setq argv (cdr argv)))

        ;; Parse the command-line options and prevent any other emacs
        ;; package from trying access the command-line arguments.
        (setq opts (ernie/parse-command-line argv))
        (setq argv nil)

        ;; Do what the user specifies in the command-line arguments.
        (cond

         ;; --help
         ((gethash :user-wants-help opts)
          (princ (format "\n%s\n" ernie/help-message)))

         ;; --version
         ((gethash :user-wants-version opts)
          (princ (apply #'format "ernie.el %d.%d.%d\n" ernie/version-info)))

         ;; Before starting to grep, make sure the user has specified
         ;; at least one pattern.
         ((not (gethash :regexps opts))
          (error "%s" "No pattern specified."))

         (t ;; Iterate over the files and directories or over stdandard input.
          (dolist (fname (or (gethash :files opts) '("-")))
            (cond ((string= fname "-")
                   (ernie/grep-standard-input "(standard input)" opts))
                  ((file-directory-p fname)
                   (if (gethash :recursive opts)
                     (ernie/dirtree fname
                                    :func #'ernie/grep-file*
                                    :initial-value opts
                                    :abspath (gethash :abspath opts))
                     (message "%s%s%s"
                              "*** Warning: "
                              "No recursion specified for directory: "
                              fname)))
                  (t (ernie/grep-file fname opts)))))))

      (error (message "*** Error: %s" (error-message-string err))
             (setq error-level 1)
             nil))

  ;; I wish there were a kinder, more gentle way of doing this, but we
  ;; need to prevent emacs from trying to process the rest of the
  ;; command line arguments.
  (kill-emacs error-level)))

(provide 'ernie)
